#include <stdio.h>
#include <string.h>
#ifdef WIN32
#include <windows.h>
#else
#include <stdlib.h>
#include <unistd.h>
#include <termios.h>
#include <sys/types.h>
#include <fcntl.h>
#define HANDLE int
#define WriteFile(A,B,C,D,E) write(A,B,C)
#define ReadFile(A,B,C,D,E) read(A,B,C)
#define FlushFileBuffers(A) fsync(A)
#define CloseHandle close
#define DWORD unsigned int
#define INVALID_HANDLE_VALUE -1
#endif

HANDLE handle;

// send NXT message string to port
void send(unsigned char *data)
{
  DWORD written;
  WriteFile(handle, data, (DWORD)(data[0] + 2), &written, 0);
  FlushFileBuffers(handle);
}

// receive one byte from port (returns after ca. 1.5 seconds if nothing received)
// returns -1 in case of error, -2 in case of timeout, otherwise the received char
int receive(void)
{
  unsigned char buf;
  DWORD nread;
  int rv = ReadFile(handle, &buf, 1, &nread, 0);
#ifdef WIN32
  if (rv == 0) return -1;
  if (nread == 0) return -2;
#else
  if (rv == -1) return -1;
  if (rv == 0) return -2;
#endif
  return buf;
}

int main(int argc, char **argv)
{
	int i;
	char portname[50];
    static unsigned char nxt_msg[30];
    unsigned char p[10];

	//extract virtual port name from command line
	if (argc != 2)
	{
		printf("usage: nxtls virtual_port_name\n");
		return 0;
	}

	int arglen = (int)strlen(argv[1]);
	if (arglen > 45) 
	{
	   printf("Do not understand %s\n", argv[1]);
	   return 1;
	}

	// strip optional column from the end of port name
	if (argv[1][4] == ':') argv[1][4] = '\0';
	if (strlen(argv[1]) > 5)
		if (argv[1][5] == ':') argv[1][5] = '\0';
	if (strlen(argv[1]) > 12)
	        if (argv[1][12] == ':') argv[1][12] = '\0';
	if (strlen(argv[1]) > 13)
	        if (argv[1][13] == ':') argv[1][13] = '\0';

#ifdef WIN32
	if ((arglen != 4) && (argv[1][0] != '\\'))
	{
		strncpy(portname, "\\\\.\\", 4);
		strncpy(portname + 4, argv[1], arglen + 1);
	}
	else 
#endif	
	  strncpy(portname, argv[1], arglen + 1);

	// open the port
#ifdef WIN32
    handle = CreateFileA(portname, FILE_READ_DATA | FILE_APPEND_DATA | FILE_WRITE_DATA, 0, NULL, OPEN_EXISTING, 0, NULL);
    if (handle != INVALID_HANDLE_VALUE)
    {
	  SetupComm(handle, 90, 90);

	  COMMTIMEOUTS timeouts;
	  timeouts.ReadIntervalTimeout = 0;
	  timeouts.ReadTotalTimeoutConstant = 0;
	  timeouts.ReadTotalTimeoutMultiplier = 1500;  //1.5s timeout on read() from serial port
	  timeouts.WriteTotalTimeoutMultiplier = 0;
	  timeouts.WriteTotalTimeoutConstant = 1500;  //1.5s timeout on write() to serial port
	  SetCommTimeouts(handle, &timeouts);

      PurgeComm(handle, PURGE_RXCLEAR | PURGE_TXCLEAR);
    }
    else
    {
	  printf("# Could not open the port, err=%d\n\n", GetLastError());
	  return 1;
    }
#else
    struct termios newtio, oldtio;
    handle = open(portname, O_RDWR | O_NOCTTY);
    if (handle < 0) { perror(portname); return INVALID_HANDLE_VALUE; }

    tcgetattr(handle,&oldtio); /* save current port settings */

    bzero(&newtio, sizeof(newtio));
    newtio.c_cflag = B115200 | CS8 | CLOCAL ;
    newtio.c_iflag = IGNPAR;
    newtio.c_oflag = 0;
    newtio.c_lflag = 0;
    newtio.c_cc[VTIME] = 0;
    newtio.c_cc[VMIN] = 1;
    cfsetspeed(&newtio, B115200);
    tcflush(handle, TCIFLUSH);
    if (tcsetattr(handle,TCSANOW,&newtio))
    {
      perror("tcsetattr()");
      close(handle);
      return 1;
    }
#endif

    // prepare NXT "find-first" message packet
    p[0] = 6;
    p[1] = 0;
    p[2] = 1; // request reply
    p[3] = 134;    // message find first
	p[4] = '*';
	p[5] = '.';
	p[6] = '*';    // ls *.*
	p[7] = 0;

    // send the prepared packet to the virtual serial port
    send(p);

    // retrieve packet with 30 bytes
    i = 0;
    while (i < 30)
    {
	  int c = receive();
	  int tm = 0;
	  while ((c == -2) && (tm++ < 5)) // timeout, but accept only 5 times
        c = receive();
      if (c < 0) return -2;

	  nxt_msg[i++] = (unsigned char)c;
    }

    unsigned long size = nxt_msg[29] * 16777216 + nxt_msg[28] * 65536 + nxt_msg[27] * 256 + nxt_msg[26];
    unsigned char ffhandle = nxt_msg[5];
	unsigned char ffstatus = nxt_msg[4];

	if (ffstatus == 135)
	{
	  printf("No files found\n");
	  CloseHandle(handle);
	  return 0;
	}

	/* if the status is error, do not continue */
	if (ffstatus)
	{
	  printf("NXT responded to find first packet with error status\n");
	  CloseHandle(handle);
	  return 1;
	}
    // print the first file
	printf("\nNXT flash memory:\n\n%s\t%d\n", nxt_msg + 6, size);

	// prepare find next packet
	p[0] = 3;
    p[1] = 0;
    p[2] = 1;      // request reply
    p[3] = 135;    // message find next
	p[4] = ffhandle;

	do {

	  // and send it
	  send(p);

      // retrieve packet with 30 bytes
      i = 0;
      while (i < 30)
      {
	    int c = receive();
	    int tm = 0;
	    while ((c == -2) && (tm++ < 5)) // timeout, but accept only 5 times
          c = receive();
        if (c < 0) return -2;

	    nxt_msg[i++] = (unsigned char)c;
      }

      size = nxt_msg[29] * 16777216 + nxt_msg[28] * 65536 + nxt_msg[27] * 256 + nxt_msg[26];
	  ffstatus = nxt_msg[4];

	  // print the next file
	  if (!ffstatus) printf("%s\t%d\n", nxt_msg + 6, size);

	} while (!ffstatus);

	/* if the status is error, do not continue */
	if (ffstatus != 135)
	{
	  printf("NXT responded to find next packet with error status\n");
      CloseHandle(handle);	
	  return 1;
	}

	// close the port
#ifdef WIN32
    PurgeComm(handle, PURGE_RXCLEAR | PURGE_TXCLEAR);
#endif
    CloseHandle(handle);

	return 0;
}
