//nxtnet.cpp - win32 console application that communicates 
//             with NXT over BT on one side and with Imagine Logo over TCP/IP on the other side

#include <winsock2.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#define MAX_RECIPIENT_LEN 10

static short debug;   // should we print debugging output
static short portOpen; // is the port successfully opened?
static HANDLE handle;  // handle to serial port
static int packetLength;   // length of the packet received
static char *packetEnd;    // end of received packet (only for decoding)

short openPort(char *port)
{
  char port2[30];
  if (strlen((char *)port) > 20) return 0;
  sprintf(port2, "\\\\.\\%s", (char *)port);

  handle = CreateFileA(port2, GENERIC_READ | GENERIC_WRITE, 0, NULL, OPEN_EXISTING, 0, NULL);
  
  if (handle != INVALID_HANDLE_VALUE) 
  {
	if (strcmp(port2, "\\\\.\\legotower1") != 0)
    {
	  COMMTIMEOUTS timeouts;
	  timeouts.ReadIntervalTimeout = 0;
	  timeouts.ReadTotalTimeoutConstant = 0;
	  timeouts.ReadTotalTimeoutMultiplier = 1500;  //1.5s timeout on read() from serial port
	  timeouts.WriteTotalTimeoutMultiplier = 0;
	  timeouts.WriteTotalTimeoutConstant = 1500;  //1.5s timeout on write() to serial port
	  SetCommTimeouts(handle, &timeouts);
  	  SetupComm(handle, 90, 90);
      PurgeComm(handle, PURGE_RXCLEAR | PURGE_TXCLEAR);
    }
    portOpen = 1;
	if (debug) printf("Opened port %s\n", port2);
  }
  else 
  {
    portOpen = 0;
	if (debug) printf("Could not open port %s, err=%d\n", port2, GetLastError());
  }

  return portOpen;
}

// closes the port
void closePort(void)
{
  if (portOpen) 
  {
	  CloseHandle(handle);
	  PurgeComm(handle, PURGE_RXCLEAR | PURGE_TXCLEAR);
      portOpen = 0;
  }
}

// send string to port
short sendToPort(char *data)
{
  DWORD written;
  short rv;
  if (!portOpen) return 0;
  //if (debug) printf("%d bytes -> NXT\n", data[0] + 2);
  if (!(rv = WriteFile(handle, data, (DWORD)(data[0] + 2), &written, 0)))
  {
    if (debug) printf("Error %d writing to port, closing port\n", GetLastError());
	CloseHandle(handle);
	portOpen = 0;
  }
  else FlushFileBuffers(handle);
  return rv;	  
}

// receive one byte from port (returns after ca. 1.5 seconds if nothing received)
short getcFromPort(char *buf)
{
  DWORD nread;
  short rv;
  if (!portOpen) return 0;
  int count = 0;
  do {
    if (portOpen) rv = ReadFile(handle, buf, 1, &nread, 0);
	if (nread == 0) 
	{
	  Sleep(15);
	  count++;
	}
	if (!rv)
	{
	  if (debug) printf("Error %d reading from serial port, closing port\n", GetLastError());
      CloseHandle(handle);
	  portOpen = 0;
	  return 0;
	}
  } while ((nread == 0) && (count < 2));
  if (count >= 2) return -1;
  return 1;
}

int sendPacket(SOCKET s, char *p, short size)
{
  int bytesSent = 0;
  
  //if (debug) printf("%d bytes -> Imagine\n", size);
  do {
    int bs = send(s, p + bytesSent, size - bytesSent, 0);
	if (bs == SOCKET_ERROR)
    {
      if (debug) printf("socket error %ld in send() while sending login confirmation, closing connection\n", WSAGetLastError());
      closesocket(s);
	  return 0;
    }
	bytesSent += bs;
  } while (bytesSent < size);

  return 1;
}

char *receivePacket(SOCKET s)
{
  char recvbuf[20];
  int delta = 0;
  int packetSize, wsaerr = 0;
  // receive packet header (<07><len><!>) 
  do {  //we have to loop for the case only part of the header will be received
	 packetSize = recv(s, ((char *)recvbuf) + delta, 1, 0);
     if (packetSize > 0) delta++;
  } while ((delta < 20) && (recvbuf[delta - 1] != '!') && (packetSize > 0));

  if (packetSize < 0) wsaerr = WSAGetLastError();

  // if connection closed, return null
  if ((packetSize == 0) || (wsaerr == WSAECONNRESET))
    return 0;

  //if error occured, report and return null
  if (packetSize < 0)
  {
	  fprintf(stderr, "error %ld reading socket\n", wsaerr);
	  return 0;
  }

  //'!' did not arrive? bad header? faulty packet
  if ((delta >= 20) || (recvbuf[0] != 7)) 
  {
	  fprintf(stderr, "Ignoring incorrect packet (%d,%c,%d)\n", delta, recvbuf[0], packetSize);
	  return 0;
  }
  
  // retrieve packet length
  recvbuf[delta - 1] = 0;
  sscanf(recvbuf + 1, "%d", &packetLength);

  //if (debug) printf("Imagine -> %d bytes\n", packetLength);
  char *packet = (char *) malloc(packetLength + 1);
  if (!packet)
  {
	  fprintf(stderr, "Could not allocate memory for packet of size %d\n", packetLength);
	  return 0;
  }
  delta = 0;
  wsaerr = 0;
  // receive packet body  
  do {  
    packetSize = ::recv(s, packet + delta, packetLength - delta, 0);
    if (packetSize > 0) delta += packetSize;
  } while ((delta < packetLength) && (packetSize > 0));

  if (packetSize < 0) wsaerr = WSAGetLastError();

  // if connection closed, return null
  if ((packetSize == 0) || (wsaerr == WSAECONNRESET))
  {
    if (debug) printf("recv(): packetSize=%d, wsaerr=%d\n", packetSize, wsaerr);
    return 0;
  }
  // if error occured, report and return null
  if (packetSize < 0)
  {
    fprintf(stderr, "error %ld while reading socket\n", wsaerr);
    return 0;
  }

  packet[packetLength] = 0;
  packetEnd = packet + packetLength;
  return packet;
}

int main(int argc, char **argv)
{
    SOCKET sock; // socket we are listening at
	int port;    // socket port number where we should be listening
	char portName[80];  // name of serial port where NXT is connected
	char packet4client[200];  // string buffer for constructing reply packet
	char recipient[MAX_RECIPIENT_LEN + 1];  // this should be only INP or OUT

	debug = 0;  // no debugging output by default
	port = 51;  // default port number for Imagine networking is 51

    if (argc > 1)
	{
      if ((_stricmp(argv[1], "-h") == 0) ||
          (_stricmp(argv[1], "-help") == 0) ||
		  (_stricmp(argv[1], "/h") == 0) ||
		  (_stricmp(argv[1], "/help") == 0))
	  {
		  printf("nxtnet [-debug] [port] - program for communication of Imagine with LEGO NXT\n");
		  return 0;
	  }
	  else if (_stricmp(argv[1], "-debug") == 0)
	  {
		  debug = 1;
	  }
	  if (argv[argc - 1][0] != '-')
	  {
		  sscanf(argv[argc - 1], "%d", &port);
	  }
	}

	if (debug) printf("nxtnet version 1.0, interfacing Imagine Logo with LEGO NXT\n");

	// initialize windows sockets
	WSADATA wsaData;
	if (int err = WSAStartup(MAKEWORD(2, 2), &wsaData))
	{
		fprintf(stderr, "Error initializing windows sockets: %d\n", err);
		return 1;
	}

	// create socket
	sock = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
    if (sock == INVALID_SOCKET)
	{
        fprintf(stderr, "Error creating socket: %ld\n", WSAGetLastError() );
        return 1;
    }

    // set up the sockaddr structure
    sockaddr_in service;
    service.sin_family = AF_INET;
	service.sin_addr.s_addr = ADDR_ANY;
	service.sin_port = htons((u_short)port);
    char enable = 1;
    setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &enable, sizeof(int));

	//bind the socket
    if (bind(sock, (SOCKADDR*) &service, sizeof(service) ) == SOCKET_ERROR)
	{
	    fprintf(stderr, "bind() failed, error=%ld.\n", WSAGetLastError() );
        closesocket(sock);
        return 1;
    }

    // listen on the socket
    if (listen(sock, 1) == SOCKET_ERROR)
	{
		fprintf(stderr, "listen() failed, error=%ld.\n", WSAGetLastError() );
        closesocket(sock);
		return 1;
	}

	// repeat forever:
	while (1)
	{
	  sockaddr_in otherAddr;
	  if (debug) printf("listening on port %d...\n", port);
      
	  // accept a new connection
	  int l = sizeof(otherAddr);
      SOCKET clientSocket = accept(sock, (SOCKADDR*) &otherAddr, &l);
 	  if (clientSocket == INVALID_SOCKET)
	  {
		int er = WSAGetLastError();
        fprintf(stderr, "accept() with error %d\n", er );

		if (er == WSAEINTR)  // this indicates that somebody killed us, terminate
		{
		  closesocket(sock);
		  return 1;
		}
	  }

	  if (debug) 
	  {
	  	char ip[100];
        strncpy(ip, inet_ntoa(otherAddr.sin_addr), 99);
	    int peerPort = ntohs(otherAddr.sin_port);
		printf("accepted connection from %s:%d\n", ip, peerPort);
	  }

	  // receive the login packet, send login reply (the name of the client is the name of the port)
      char *p = receivePacket(clientSocket);
	  if (!p)
	  {
		if (debug) printf("Error reading login packet from the client, closing connection.\n");
		closesocket(clientSocket);
		continue;
	  }

	  // verify the login packet
	  if ((strlen(p) < 4) || (p[0] !='a') || (p[2] != '!')) 
	  {
		if (debug) printf("Incorrect login packet from client, closing connection.\n");
		closesocket(clientSocket);
		continue;
	  }
	  // save the client name
	  strncpy(portName, p + 3, 79);
	  portName[79] = 0;
	  // and release the login packet
	  free(p);

	  // send the login confirmation packet to the client
	  sprintf(packet4client, "\007%d!b0!%s", strlen(portName) + 3, portName);
      if (!sendPacket(clientSocket, packet4client, strlen(packet4client))) continue;

	  // open the port
	  openPort(portName);

	  // repeat until the client disconnects
	  while (1)
	  {
	    // wait for a packet from the client
	    p = receivePacket(clientSocket);

		if (!p) // connection lost?
		{
		  if (debug) printf("connection with the client lost.\n");
		  closesocket(clientSocket);
		  break;
		}

		// verify the packet
		if ((strlen(p) < 3))
		{
		  if (debug) printf("incorrect packet from client received, closing connection\n");
		  free(p);
		  closesocket(clientSocket);
		  break;
		}

		// the list of users message 
		if (p[0] == 'U') // send back list of users
		{
		  sprintf(packet4client, "\007%d!v2!3!NXT%d!%s", strlen(portName) + 10, strlen(portName), portName);
		  if (!sendPacket(clientSocket, packet4client, strlen(packet4client)))
		  {
		    if (debug) printf("error sending list of users packet to client, closing connection\n");
			free(p);
			break;
		  }
		}
		else if (p[0] == 'L') // list message is used only to clear the message queue
		{
		  // send the empty list packet back to the client
          if (!sendPacket(clientSocket, "\0078!l1!3!NXT", 11)) 
		  {
		    if (debug) printf("error sending list message back to client, closing connection\n"); 
	        free(p);
			break;
		  }
		}
		else if ((toupper(p[0]) == 'F') || (toupper(p[0]) == 'G')) // ignore getinfo-packages for now
		{
		}
		else if (p[0] == 'W') // regular message
		{
		  char *messageData;
		  int l1;
		  //client name length
		  char *t = strtok(p + 3, "!");
		  if (t) sscanf(t, "%d", &l1);
		  //recipient name length
		  t = strtok(0, "!");
		  if (t) sscanf(t + l1, "%d", &l1);
		  //data
		  t = strtok(0, "!");
  		  //determine recipient name
		  if (t) 
		  {
		    int rd = (l1>MAX_RECIPIENT_LEN)?MAX_RECIPIENT_LEN:l1;
			strncpy(recipient, t, rd);
			recipient[rd] = '\0';
		    messageData = t + l1;
		  }
		  else 
		  {
            if (debug) printf("incorrectly formatted W packet, closing connection\n");
			free(p);
			closesocket(clientSocket);
		    break;
		  }

		  // port closed? attempt to re-open
		  if (!portOpen) openPort(portName);

		  if (_stricmp(recipient, "out") == 0)  // send request
		  {
		    // message is encoded (chars with ascii < 48 are sent as #48#(a+48)
		    char *md = (char *)malloc(packetLength + 1), *rp = messageData;
			char *wp = md;
			while (rp < packetEnd)
			{
				if (*rp == '0') 
				{
					rp++;
					*wp = (*rp) - 48;
				}
				else *wp = *rp;
				rp++;
				wp++;
			}

		    // write to port - send the data further down to the port, and send back status
            if (sendToPort(md))
			{
			  free(md);
			  // send the confirmation packet to the client
              if (!sendPacket(clientSocket, "\00710!w1!3!NXTOK", 14)) 
			  {
			    free(p);
			    break;
			  }
			}
			else // sending not successful
			{
			  free(md);
			  // send the error packet to the client
              if (!sendPacket(clientSocket, "\00711!w1!3!NXTERR", 15)) 
			  {
			    free(p);
			    break;
			  }
			}
		  }
		  else if (strcmp(recipient, "INP") == 0) // read request
		  {
		    // read from port - read data from port and send them back
			
			// how many bytes to read?
			sscanf(messageData, "%d", &l1);
			int toRead = l1;

			char *buf = (char *)malloc((l1<3)?4:(l1+1));
			if (!buf)
			{
				fprintf(stderr, "Could not allocate memory for input packet of size %d\n", l1);
				free(p);
				break;
			}
			char *b = buf;

			while (portOpen && (l1--))
			{
			  switch (getcFromPort(b))
			  {
			    case 1: b++;
					    break;
				case 0: strcpy(buf, "ERR");
					    b = buf + 3;
						toRead = 3;
					    l1 = 0;
					    break;
				case -1: strcpy(buf, "TO");
					     b = buf + 2;
						 toRead = 2;
					     l1 = 0;
						 break;
			  }
			}
			*b = '\0';
			char *pk = (char *)malloc(toRead + 20);
			if (!pk)
			{
				fprintf(stderr, "Could not allocate input packet of size %d\n", toRead + 20);
				free(p);
				free(buf);
				break;
			}

			//if (debug) printf("NXT -> %d bytes\n", toRead);

			// send the data read from port to the client
			sprintf(pk, "\007%d!w1!3!NXT", 8 + toRead);
			l1 = strlen(pk);
			memcpy(pk + l1, buf, toRead);
			free(buf);
            if (!sendPacket(clientSocket, pk, toRead + l1)) 
			{
			  free(p);
			  free(pk);
			  break;
			}
			free(pk);
		  }
		  else
		  {
		    if (debug) printf("Unrecognized recipient %s\n", recipient);
		  }
		} // 'W' - regular message
		else 
		{
		  if (debug) printf("unexpected packet '%c' received and ignored.\n", p[0]);
		}
		// release the packet
		free(p);
	  } // while endless loop for this client

      // if we came here, the connection with the client has been closed

      // close the port
	  closePort();
	
    } // while(1)

	return 0; // never reached
} // main()
