// setup.cpp - win32 console program for setting up NXT Logo

#include <stdio.h>
#include <string.h>
#include <errno.h>
#ifdef WIN32
#include <windows.h>
#include <direct.h>
char slash='\\';
char *prefix = "";
#else
#include <stdlib.h>
#include <unistd.h>
#include <termios.h>
#include <sys/types.h>
#include <fcntl.h>
#define HANDLE int
#define WriteFile(A,B,C,D,E) write(A,B,C)
#define ReadFile(A,B,C,D,E) read(A,B,C)
#define FlushFileBuffers(A) fsync(A)
#define CloseHandle close
#define DWORD unsigned int
#define INVALID_HANDLE_VALUE -1
#define _mkdir(X) mkdir(X,0770)
#define Sleep(X) usleep(1000*X)
char slash='/';
char *prefix = "./";
#endif

HANDLE handle;

// send NXT message string to port
void send(unsigned char *data)
{
  DWORD written;
  WriteFile(handle, data, (DWORD)(data[0] + 2), &written, 0);
  FlushFileBuffers(handle);
}

// receive one byte from port (returns after ca. 1.5 seconds if nothing received)
// returns -1 in case of error, -2 in case of timeout, otherwise the received char
int receive(void)
{
  unsigned char buf;
  DWORD nread;
  int rv = ReadFile(handle, &buf, 1, &nread, 0);
#ifdef WIN32
  if (rv == 0) return -1;
  if (nread == 0) return -2;
#else
  if (rv == -1) return -1;
  if (rv == 0) return -2;
#endif
  return buf;
}

char *nxt_error(int e)
{
	static char *nxterrs[] = { "No more handles", "No space", "No more files",
		                       "End of file expected", "End of file", "Not a linear file",
							   "File not found", "Handle all ready closed", "No linear space",
							   "Undefined error", "File is busy", "No write buffers",
							   "Append not possible", "File is full", "File exists",
							   "Module not found", "Out of boundary", "Illegal file name",
							   "Illegal handle" };
	if (!e) return "OK";
	if ((e > 0x80) && (e <= 0x93)) return nxterrs[e - 0x81];
	return "Unknown Error";
}

int main()
{
	char portname[100], portn[100], portname0[20];
    unsigned char nxt_msg[100];
	unsigned char p[100];
	int i;

	printf("\nWelcome to NXT Logo Setup program\n\n");
	printf("Before continuing, it is strongly recommended that you install the\n");
	printf("extended firmware in your NXT brick. Extended firmware allows much\n");
	printf("faster operation of NXT Logo and leaves much more memory for your\n");
	printf("Logo programs. It is compatible with default firmware, so you can\n");
	printf("continue using your NXT with the extended firmware as before...\n");
	printf("You can download it from John Hansen's BricX website here:\n");
	printf(" http://bricxcc.sourceforge.net/lms_arm_jch.zip\n");
	printf("If you do not want to use the extended firmware, you must rename\n");
	printf("the file logo_standard.rxe to logo.rxe before continuing.\n");
	printf("If you did not install extended firmware yet, or renamed the file,\n");
	printf("please exit the setup program now, and return back when ready.\n");
	printf("\nAre you ready to proceed with setup? [yes/no]: ");
	fgets(portname, 10, stdin);
	if ((portname[0] != 'Y') && (portname[0] != 'y')) return 0;

	printf("------------------------------------------------------------------\n");
	printf("Now I need to learn your outgoing virtual serial port for\n");
	printf("PC -> NXT connection.\n\n");
	printf("If you do not have one yet, insert BT dongle into free USB port,\n");
	printf("wait until it is properly recognized and installed, turn on your\n");
	printf("NXT brick, configure the BlueTooth on your NXT to be ON and VISIBLE.\n");
#ifdef WIN32
	printf("Then double click on the BlueTooth icon in the system tray (if you do\n");
	printf("not see it, you find it in the Control Panel), and add the NXT device.\n");
	printf("This will create both incoming and outgoing virtual serial ports.\n");
	printf("The port name should be in the format COMX or COMXX: COM7, COM17, etc.\n");
	char *exaport = "COM17";
#else
	printf("Then search for NXT device using command 'hcitool scan' and note the\n");
	printf("BT address. In a separate shell, connect NXT brick using\n");
	printf("'rfcomm connect 0 BT_ADDRESS' thus NXT will be connected to /dev/rfcomm0\n");
	printf("During this process you will need to provide the PIN on both sides.\n");
	char *exaport = "/dev/rfcomm0";
#endif
    printf("------------------------------------------------------------------\n");
	printf("Your outgoing NXT virtual serial port (ex. %s): ", exaport);
    fgets(portn, 99, stdin);
	portn[strlen(portn) - 1] = '\0';
	printf("Verifying connection...\n");

	int arglen = (int)strlen(portn);

	strncpy(portname0, portn, 19);
	portname0[19] = 0;

#ifdef WIN32
	if ((arglen != 4) && (portn[0] != '\\'))
	{
		strncpy(portname, "\\\\.\\", 4);
		strncpy(portname + 4, portn, arglen + 1);
	}
	else 
#endif
	  strncpy(portname, portn, arglen + 1);

	// open the port
#ifdef WIN32
    handle = CreateFileA(portname, FILE_READ_DATA | FILE_APPEND_DATA | FILE_WRITE_DATA, 0, NULL, OPEN_EXISTING, 0, NULL);
    if (handle != INVALID_HANDLE_VALUE) 
    {
	  SetupComm(handle, 90, 90);

	  COMMTIMEOUTS timeouts;
	  timeouts.ReadIntervalTimeout = 0;
	  timeouts.ReadTotalTimeoutConstant = 0;
	  timeouts.ReadTotalTimeoutMultiplier = 1500;  //1.5s timeout on read() from serial port
	  timeouts.WriteTotalTimeoutMultiplier = 0;
	  timeouts.WriteTotalTimeoutConstant = 1500;  //1.5s timeout on write() to serial port
	  SetCommTimeouts(handle, &timeouts);

      PurgeComm(handle, PURGE_RXCLEAR | PURGE_TXCLEAR);
    }
    else 
    {
	  printf("Could not open the port, err=%d\n\n", GetLastError());
 	  printf("Press ENTER to quit the setup.\n");
	  fgets(portn, 10, stdin);
	  return 1;
    }
#else
    struct termios newtio, oldtio;
    handle = open(portname, O_RDWR | O_NOCTTY);
    if (handle < 0) { perror(portname); return INVALID_HANDLE_VALUE; }

    tcgetattr(handle,&oldtio); /* save current port settings */

    bzero(&newtio, sizeof(newtio));
    newtio.c_cflag = B115200 | CS8 | CLOCAL ;
    newtio.c_iflag = IGNPAR;
    newtio.c_oflag = 0;
    newtio.c_lflag = 0;
    newtio.c_cc[VTIME] = 0;
    newtio.c_cc[VMIN] = 1;
    cfsetspeed(&newtio, B115200);
    tcflush(handle, TCIFLUSH);
    if (tcsetattr(handle,TCSANOW,&newtio))
    {
      perror("tcsetattr()");
      close(handle);
      return 1;
    }
#endif

    // get firwmare version packet
	p[0] = 2;
	p[1] = 0;
	p[2] = 1;
	p[3] = 0x88;  
	send(p);

	//check if write was successful
	i = 0;
	while (i < 9)
    {
	  int c = receive();
	  int tm = 0;
	  while ((c == -2) && (tm++ < 5))   // timeout, but accept only 5 times
        c = receive();
      if (c < 0) 
	  {
		printf("Error getting answer from NXT\n");
		CloseHandle(handle);
		printf("Press ENTER to quit the setup.\n");
	    fgets(portn, 10, stdin);
		return 1;
	  }
	  nxt_msg[i++] = (unsigned char)c;
    }

	// is connection ok?
    if (nxt_msg[4])
	{
		printf("Error getting answer from NXT, err: %s\n", nxt_error(nxt_msg[4]));
		CloseHandle(handle);
		printf("Press ENTER to quit the setup.\n");
	    fgets(portn, 10, stdin);
		return 1;
    }

	printf("OK, firmware version %d.%d, protocol version %d.%d\n", 
		nxt_msg[8], nxt_msg[7], nxt_msg[6], nxt_msg[5]);

    // send two beeps
	p[0] = 6;
	p[1] = 0;
	p[2] = 0x80;
	p[3] = 3;
	p[4] = 2640 & 0xFF;
	p[5] = (2640 >> 8);
	p[6] = 50;
	p[7] = 0;
	send(p);
	Sleep(100);
	send(p);

	// and close the connection to NXT
	CloseHandle(handle);

	printf("------------------------------------------------------------------\n");
	printf("NXT Logo interpreter requires enough free space in the NXT flash.\n");
	printf("Therefore, I will now take a backup of all your files in NXT flash\n");
	printf("to directory nxt_flash_backup%c and then try to remove all unnecessary\n", slash);
	printf("files. You can recover the files anytime using the nxtcp utility, or\n");
	printf("by reinstalling the default NXT firmware, which will also copy all\n");
	printf("standard sound, image, and try* files. OK to continue? [yes/no]: ");
    fgets(portn, 10, stdin);
	if ((portn[0] != 'y') && (portn[0] != 'Y')) return 0;
	
	// create backup directory
	if (_mkdir("nxt_flash_backup"))
	  if (errno != EEXIST)
	  {
	    printf("Could not create directory nxt_flash_backup, errno=%d\n", errno);
		printf("Press ENTER to quit the setup.\n");
	    fgets(portn, 10, stdin);
	    return 1;
	  }
	
    // backup all files from NXT Flash
	printf("\nCopying all files from NXT flash...\n");
	sprintf(portn, "%snxtcp %s:*.* nxt_flash_backup%c", prefix, portname0, slash);
	if (-1 == system(portn))
	{
		printf("Could not execute nxtcp utility, is it missing? errno=%d\n", errno);
		printf("Press ENTER to quit the setup.\n");
	    fgets(portn, 10, stdin);
		return 1;
	}
	Sleep(3000);  // these waits are necessary for the windows BT stack to recover... :(

	// remove all files from NXT flash
	printf("\nRemoving unnecessary files...\n");
	sprintf(portn, "%snxtrm %s:*.*", prefix, portname0);
	if (-1 == system(portn))
	{
		printf("Could not execute nxtrm utility, is it missing? errno=%d\n", errno);
		printf("Press ENTER to quit the setup.\n");
	    fgets(portn, 10, stdin);
		return 1;
	}
	Sleep(3000);

	// copy required NXT Logo files
	printf("\nInstalling NXT Logo files...\n");
    sprintf(portn, "%snxtcp reslen.dat %s:", prefix, portname0);
	if (-1 == system(portn))
	{
		printf("Could not execute nxtcp utility, is it missing? errno=%d\n", errno);
		printf("Press ENTER to quit the setup.\n");
	    fgets(portn, 10, stdin);
		return 1;
	}
    Sleep(3000);

	sprintf(portn, "%snxtcp reserved.dat %s:", prefix, portname0);
	if (-1 == system(portn))
	{
		printf("Could not execute nxtcp utility, is it missing? errno=%d\n", errno);
		printf("Press ENTER to quit the setup.\n");
	    fgets(portn, 10, stdin);
        return 1;
	}
    Sleep(3000);
	
	sprintf(portn, "%snxtcp logo.rxe %s:", prefix, portname0);
	if (-1 == system(portn))
	{
		printf("Could not execute nxtcp utility, is it missing? errno=%d\n", errno);
		printf("Press ENTER to quit the setup.\n");
	    fgets(portn, 10, stdin);
        return 1;
	}

	//complete
	printf("------------------------------------------------------------------\n");
    printf("NXT Logo is ready for use. Either load the");
#ifdef WIN32
    printf(" nxt.imp project from Imagine\n");
#else
    printf(" nxt.lgo project from UCBLogo\n");
#endif
	printf("or run './remote %s' from command prompt. Please refer to the user\n", portname0);
	printf("and reference guides for more information. NXT Logo home is:\n");
	printf("  http://robotics.sk/NXTLogo/\n");
    printf("------------------------------------------------------------------\n\n");
	printf("Press ENTER to quit the setup.\n");
	fgets(portn, 10, stdin);

	return 0;
}
